/**
 * @file   nu_convexity.c
 * @author David Bremner <bremner@unb.ca>
 * @date   Mon Dec 30 18:21:28 2002
 * 
 * @brief  Manage convexity constraints 
 * 
 */

#include "nalloc.h"
#include "basis_equation.h"
#include "combinat.h"

/** 
 * allocate a block for convexity equations.
 * 
 * @param state node to hold block
 * @param n elements
 * @param r rank
 * 
 * @return modified node
 */
na_node_p ce_nalloc(na_node_p state,slot_t slot, int  n, int r){
  int rows=binomial(n,r)*(n-r);
  int cols=r+1;

  state=be_nalloc(state,slot,rows,cols,binomial(n,r));
  return state;
}


/** 
 * initialize a set of constraints for checking Carotheodory's condition
 * for interior points.
 * 
 * @param bdy_eq preallocated block
 * @param n elements
 * @param r rank
 *
 * The idea is to check for each vertex $v$, for each simplex T (d+1 points) 
 * is $v$ in the interior of $T$.  This is equivalent to checking for 
 * each facet F of T, is $v$ on the same side of $F$ as $T\F$, i.e. of $t=T\F$
 * $[ F t ] = [ F v ]$. Since all of left hand sides are just permutations
 * of [ T ], we can reduce ask if [ T ] * sigma_0 , [ F v ] * sigma_F have
 * the same signs, for appropriate constants $sigma_i$.
 */
void ce_init(be_block_p conv_eq, int n, int r){

  natural_t *pi,*pibar,*tau;
  array_index_t i;
  int row=0;
  conv_eq->stride=1;

  pi=calloc(r,sizeof(natural_t));
  pibar=calloc(n-r,sizeof(natural_t));
  tau=calloc(n-r,sizeof(natural_t));

  first_kpartition(n,r,pi,pibar,tau);

  /* iterate over r-simplices */
  do {
    /* for each vertex not in that simplex, make a row to test if it
       is interior */

    
    for (i=0; i<n-r; i++){
      int j;
      bitmask_t rmask=~0x1U;

      array_index_t vertex=pibar[i];
    
      be_set_sign(conv_eq,row,0,1);
      be_set_bindex(conv_eq,row,0,kset_index(n,r,pi));

      for (j=0 ;j<r ; j++){
	int sigma=tau[i];
	rmask <<= 1;

	/* if moving that j'th to the end of the bracket crosses an odd number
	   of elements */
 	if ( (r-j-1) %2 ) 
 	  sigma *=-1; 

	/* one more swap removed if we would have had to move over 
	   the element we removed */

	if (vertex<pi[j])
	  sigma *= -1;

	be_set_sign(conv_eq,row,j+1,sigma);

	swap(pi[j],vertex);
	be_set_bindex(conv_eq,row,j+1,kset_cindex(n,r,pi));
	swap(pi[j],vertex);
      }
      conv_eq->row_mask[row]=~rmask;
      row++;
    }
  } while(next_kpartition(n,r,pi,pibar,tau));

  free(pi);
  free(pibar);
  free(tau);
  be_makelinks(conv_eq);
}


