#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include "oms_error.h"
#include "oms_types.h"
#include "scan.h"
#include "basis_equation.h"
#include "eqout.h"

static void usage(char *name){
  printf("usage: %s [-o outfile ] -d dim -k steps -n elements\n",name);
  exit(1);
}

/**
   generate constraints for a single path.

   Note that this relies on the rather special form of paths
   compared to random sets of facets.  Each facet
   is interpreted as a set of basis signs that are equal iff the facet
   exists. Sets corresponding to successive facets on the path
   overlap by exactly one basis sign. No other facet-sets on the
   path overlap.
*/
  
static void process_path(int *path, int len, be_block_p conv_eq){
  int i,j;
  int negs=0;
  int *last_sign;
  int group_sign=1;

  last_sign=calloc(be_bases(conv_eq)+1,sizeof(int));

  eq_zero();

  for (i=0; i< len; i++){

    /* pass 1, decide how the inequalities for this facet are coupled to 
       the previous one 
    */
    for (j=0; j< be_cols(conv_eq); j++){
      int b=be_bindex(conv_eq,path[i],j);
      int sigma=be_sign(conv_eq,path[i],j);
      if (sigma==-1*last_sign[b]){
	group_sign *=-1;
      }
      last_sign[b]=sigma;
    }

    for (j=0; j< be_cols(conv_eq); j++){
      if (group_sign*be_sign(conv_eq,path[i],j)==1){
	eq_coeff_add(be_bindex(conv_eq,path[i],j),1);
      } else {
	eq_coeff_add(be_bindex(conv_eq,path[i],j),-1);
	negs++;
      }
    }
  }

  eq_rhs(len*(be_cols(conv_eq))-1-negs);

  eq_type(LEQ);
  eq_output();

  eq_rhs(1-negs);
  eq_type(GEQ);
  eq_output();
  free(last_sign);
}

int main(int argc, char **argv){

  int *path;

  int c;
  int elements=-1,dim=-1,steps=-1,facets;
  be_block_p conv_eq;
  int outform=CPLEX;
  int *row;

  while((c=getopt(argc,argv,"i:o:d:k:n:f:"))!=-1){
    switch(c){
    case 'd':
      dim=atoi(optarg);
      break;
    case 'k':
      steps=atoi(optarg);
      break;
    case 'n':
      elements=atoi(optarg);
      break;
    case 'o':
      freopen(optarg,"w",stdout);
      oms_assert(stdout);
      break;
    case 'i':
      freopen(optarg,"r",stdin);
      oms_assert(stdin);
      break;
    case 'f':
      outform=eq_arg(optarg);
      break;

    default:
      usage(argv[0]);
      exit(1);
    }
  }
  
  eq_init(outform,binomial(elements,dim+1));
  facets=binomial(elements,dim);

  if (elements==-1 || dim==-1)
    oms_fatal(HERE,"you must specify dimension and number of elements");

  if (steps==-1){
    steps=elements-dim;
  }
  conv_eq=be_alloc(facets,(elements-dim), binomial(elements,dim+1));
  be_boundary_init(conv_eq,elements,dim+1,0,facets,-1);

  row=calloc(steps+1,sizeof(int));

  while(1){
    int count=0;
    int tok;
    while (count < steps && ((tok=scanlex()) == SCAN_INT)  ){
      row[count++]=scanint;
    }
    if (tok== SCAN_EOF)
      break;
    process_path(row,count,conv_eq);
  }
  return(0); 
}
