/* This file contains routines to deal with the diameter of 
 * the ridge graph.  Since we are given a diametral path, all
 * we need to do is ensure that no shortcuts are created between 
 * the initial and terminal facet.
 */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include "basis_equation.h"
#include "bindex_table.h"
#include "combinat.h"
#include "debug.h"
#include "int_queue.h"
#include "int_stack.h"
#include "oms_opt.h"
#include "profile.h"
#include "rg_internal.h"
#include "statearray.h"

TAG_DECLARE(rg_fixup);


/** 
 * allocate an empty ridge graph.
 * 
 * @param node		memory block
 * @param slot		what slot to occupy
 * @param n		how many elements
 * @param r		the rank of the oriented matroid
 * @param source	source facet
 * @param sink		sink facet
 * @param diameter	distance to maintain between source and sink
 * @return		the updated na_node
 */
na_node_p rg_nalloc(na_node_p node, 
		    slot_t slot, 
		    natural_t n, natural_t r,
		    array_index_t source, array_index_t sink, natural_t diameter){
  natural_t bases,d=r-1,i;
  ridge_graph RG;

  bases=binomial(n,d);

  node=na_alloc(node,slot, sizeof(ridge_graph_t));
  na_set_fixup(node,slot,TAG(rg_fixup));

  /** \todo the stacksize parameters here are bogus and will
     probably crash if VAR_STACK is disabled.
  */
  node=sa_nalloc(node,slot+RG_OFF_source_dist,bases,bases);
  node=sa_nalloc(node,slot+RG_OFF_sink_dist,bases,bases);

  RG=na_access(node,slot);

  RG->dim=d;
  RG->elements=n;
  RG->bases=bases;
  RG->source=source;
  RG->sink=sink;
  RG->diameter=diameter;

  rg_fixup(node,slot);
  sa_fixup(node,slot+RG_OFF_source_dist);
  sa_fixup(node,slot+RG_OFF_sink_dist);

  for (i=0; i< bases; i++){
    sa_init(RG->source_dist,i,RG_INFINITY);
    sa_init(RG->sink_dist,i,RG_INFINITY);
  }

  if (source>=0)
    sa_init(RG->source_dist,source,0);
  if (sink>=0)
    sa_init(RG->sink_dist,sink,0);
  return node;
}

/** 
 * fixup any pointers in a ridge graph
 * 
 * @param node memory block
 * @param slot base slot
 */
   
void rg_fixup(na_node_p node, slot_t slot){
  ridge_graph_p RG;

  /* note, the sarrays should be fixed up 
     by their own hooks */

  RG=na_access(node,slot);
  RG->source_dist=na_access(node,slot+RG_OFF_source_dist);
  RG->sink_dist=na_access(node,slot+RG_OFF_sink_dist);
}

/**
   \todo Why is this not automatic ?
*/
#if 0
void rg_forget(na_node_p node, slot_t slot){
  sa_forget(node,slot+RG_OFF_source_dist);
  sa_forget(node,slot+RG_OFF_sink_dist);
}
#endif

/** 
 * Save state, to be restored with rg_restore()
 * 
 * @param G ridge graph
 */
void rg_save(ridge_graph G){
  err_assert(G->dim>0);

  sa_save(G->source_dist);
  sa_save(G->sink_dist);
}

/** 
 * restore state saved by rg_save()
 * 
 * @param G ridge graph
 */
void rg_restore(ridge_graph G){

  sa_restore(G->source_dist);
  sa_restore(G->sink_dist);
}



static bool rg_check_diameter(ridge_graph RG){
  
  int delta;
  
  if (RG->sink== -1 || RG->source==-1 || RG->diameter==-1)
    return TRUE;

  delta=sa_read(RG->source_dist,RG->sink);

  return(delta>=RG->diameter);
}

/** 
 * add a facet (i.e. a node of the ridge graph).
 * 
 * @param G ridge graph
 * @param new_facet index of facet to be added
 * @param boundary state of boundary, i.e. what facets must, or must not be on
 * @param pivots lookup table of possible neighbours
 * 
 * @return true if successful
 */
bool rg_add_facet(ridge_graph G, 
		  array_index_t new_facet, 
		  sarray_p boundary,
		  be_block_p pivots
		  ){
  err_debug(DB_RIDGEGRAPH,(HERE,"adding facet %d",new_facet));
  err_assert(sa_read(boundary,new_facet)==1);

  if (!update_distances(G,new_facet,G->source_dist,G->sink_dist,boundary,pivots))
    ERR_RETURN(FALSE);

  if (!update_distances(G,new_facet,G->sink_dist,G->source_dist,boundary,pivots))
    ERR_RETURN(FALSE);

  return(rg_check_diameter(G));
}
  





/** 
 * Update the distances from an (implicit) source after adding a single
 * new_facet.
 *
 * @param RG the ridge graph
 * @param new_facet facet to start from
 * @param distances to update
 * @param other_distances coming from other of source/sink
 * @param boundary which facets are on boundary
 * @return false if we create a shortcut, true otherwise
 */
static bool update_distances(ridge_graph RG, 
			     array_index_t new_facet, 
			     sarray_p distance,
			     sarray_p other_distance,
			     sarray_p boundary,
			     be_block_p pivots){

  int i,new_dist;
  int_queue_p facet_Q;

  /* first, lets get the distance to the new facet correct */

  if (sa_read(distance,new_facet)==0){
    /* we are inserting a source or sink */
    new_dist=0;
  } else {

    new_dist=RG_INFINITY;
    
    for (i=0; i<be_cols(pivots); i++){
      int nbr=be_bindex(pivots,new_facet,i);
      if (sa_read(boundary,nbr)>0){
	int dist=sa_read(distance,nbr);
	err_assert(dist>=0);
	
	if (dist < (new_dist-1)){
	  new_dist=dist+1;
	} 
      }
    }
  }
  if (new_dist < RG_INFINITY ){
    natural_t odist=sa_read(other_distance,new_facet);

    if ((odist<RG_INFINITY) &&  (new_dist+odist < RG->diameter) )
      return FALSE;

    sa_set(distance,new_facet,new_dist);


    /* now do (directed) BFS */
    
    facet_Q=iq_new(RG->bases);
    iq_enq(facet_Q,new_facet);

    while (!iq_empty(facet_Q)){
      int parent,parent_dist;
      
      parent=iq_deq(facet_Q);
      parent_dist=sa_read(distance,parent);
    
      for (i=0; i<be_cols(pivots); i++){
	array_index_t nbr=be_bindex(pivots,parent,i);
	natural_t dist=sa_read(distance,nbr);
	natural_t new_dist=parent_dist+1;

	if (dist > new_dist){
	  odist=sa_read(other_distance,nbr);

	  if ((odist<RG_INFINITY) && ((odist+new_dist)<RG->diameter)){
	    int bd_state=sa_read(boundary,nbr);
	    if (bd_state>0){
	      iq_destroy(facet_Q);
	      return FALSE;
	    } else if (bd_state==0){
	      sa_set(boundary,nbr,-1); 
	      maybe_prof_event(PROF_UNBOUNDARY_FORCED);
	      err_debug(DB_RIDGEGRAPH|DB_FORCE,(HERE,"forbidding facet %d",nbr)); 
	    }
		
	  } else {
	    sa_set(distance,nbr,new_dist);
	    if (sa_read(boundary,nbr)>0) 
	      iq_enq(facet_Q,nbr);
	  } 

	}
      }
    }

    iq_destroy(facet_Q);
  }  else {
    err_debug(DB_RIDGEGRAPH,(HERE,"disconnected facet %d",new_facet));
  }
  return TRUE;
}




#if 0

/** 
 * dump the adjacencies of a ridge graph
 * 
 * @param G ridge graph
 */
extern void rg_dump(ridge_graph G){
  array_index_t i,j;
  for (i=0 ; i< rg_facets(G); i++){
    printf("%d:\t",rg_lookup_basis(G,i));
    for (j=0; j< rg_rcount(G,i); j++){
      printf("%d\t",rg_neighbour(G,i,j));
    }
    printf("\n");
  }   
}

#endif
